"""
This module contains the classes and functions for creating demographics files
for HIV simulations. For more information on |EMOD_s| demographics files,
see :doc:`emod-hiv:software-demographics`. 
"""

import emod_api.demographics.Demographics as Demog
from emod_api.demographics import DemographicsTemplates as DT
from emodpy_hiv.demographics import DemographicsTemplates as HIV_DT

# the mortality algo has a few options. Might want to make them public module variables
# rather than fill up the function with them.
add_to_list=True

class HIVDemographics(Demog.Demographics):
    """
    This class is derived from :py:class:`emod_api:emod_api.demographics.Demographics.Demographics` 
    and sets certain defaults for HIV in construction.

    Args:
        nodes: The number of nodes to create.
        idref: Method describing how the latitude and longitude values are created
            for each of the nodes in a simulation. "Gridded world" values use a grid 
            overlaid across the globe at some arcsec resolution. You may also generate 
            the grid using another tool or coordinate system. For more information,
            see :ref:`emod-hiv:demo-metadata`.
        base_file: A basic demographics file used as a starting point for
            creating more complicated demographics files. For example,
            using a single node file to create a multi-node file for spatial
            simulations. 
        init_prev: The initial HIV prevalence of the population.

    Returns: 
        None 
     """
    def __init__(self, nodes, idref="Gridded world grump2.5arcmin", base_file=None ):
        super().__init__( nodes, idref, base_file )
        #super().SetDefaultProperties()
        super().SetDefaultNodeAttributes(birth=True)
        DT.FullRisk( self )
        HIV_DT.AddDefaultSociety( self )

    def fertility( self, path_to_csv ):
        """
        Set fertility based on data. Simulation shall concist of individual pregnancies with rates by 
        woman's age and year-of-simulation using data from provided csv.
        """
        super().SetDefaultFromTemplate( DT.get_fert_dist( path_to_csv ), DT._set_fertility_age_year )

    def mortality(
            self,
            file_male,
            file_female,
            interval_fit=[1970, 1980],
            which_point='mid',
            predict_horizon=2050,
            csv_out=False,
            n=0, # I don't know what this means
            results_scale_factor=1.0/365.0):
        """
        Calculate and set the expected natural mortality by age, sex, and year from data, predicting what it would
        have been without HIV.
        """
        from collections import OrderedDict
        from sklearn.linear_model import LinearRegression
        from functools import reduce

        name_conversion_dict = {'Age (x)': 'Age',
                                'Central death rate m(x,n)': 'Mortality_mid',
                                'Age interval (n)': 'Interval',
                                'Period': 'Years'
                                }
        sex_dict = {'Male': 0, 'Female': 1}

        def construct_interval(x, y):
            return (x, x + y)

        def midpoint(x, y):
            return (x + y) / 2.0

        def generate_dict_order(tuple_list, which_entry=1):
            my_unordered_list = tuple_list.apply(lambda x: x[which_entry])
            dict_to_order = OrderedDict(zip(tuple_list, my_unordered_list))
            return dict_to_order

        def map_year(x_tuple, flag='mid'):
            valid_entries_loc = ['mid', 'end', 'start']

            if flag not in valid_entries_loc:
                raise ValueError('invalid endpoint specified')

            if flag == 'mid':
                return (x_tuple[0] + x_tuple[1]) / 2.0
            elif flag == 'start':
                return x_tuple[0]
            else:
                return x_tuple[1]

        import pandas as pd
        import numpy as np
        df_mort_male = pd.read_csv(file_male, usecols=name_conversion_dict)
        df_mort_male['Sex'] = 'Male'
        df_mort_female = pd.read_csv(file_female, usecols=name_conversion_dict)
        df_mort_female['Sex'] = 'Female'
        df_mort = pd.concat([df_mort_male, df_mort_female], axis=0)
        df_mort.rename(columns=name_conversion_dict, inplace=True)
        df_mort['Years'] = df_mort['Years'].apply(lambda xx: tuple(
            [float(zz) for zz in xx.split('-')]))  # this might be a bit too format specific (ie dashes in input)

        # log transform the data and drop unneeded columns
        df_mort['log_Mortality_mid'] = df_mort['Mortality_mid'].apply(lambda x: np.log(x))
        df_mort['Age'] = df_mort[['Age', 'Interval']].apply(lambda zz: construct_interval(*zz), axis=1)

        year_order_dict = generate_dict_order(df_mort['Years'])
        age_order_dict = generate_dict_order(df_mort['Age'])
        df_mort['sortby2'] = df_mort['Age'].map(age_order_dict)
        df_mort['sortby1'] = df_mort['Sex'].map(sex_dict)
        df_mort['sortby3'] = df_mort['Years'].map(year_order_dict)
        df_mort.sort_values(['sortby1', 'sortby2', 'sortby3'], inplace=True)
        df_mort.drop(columns=['Mortality_mid', 'Interval', 'sortby1', 'sortby2', 'sortby3'], inplace=True)

        # convert to years (and to string for age_list due to really annoying practical slicing reasons
        df_mort['Years'] = df_mort['Years'].apply(lambda x: map_year(x, which_point))
        df_mort['Age'] = df_mort['Age'].apply(lambda x: str(x))
        df_before_time = df_mort[df_mort['Years'].between(0, interval_fit[0])].copy()

        df_mort.set_index(['Sex', 'Age'], inplace=True)
        sex_list = list(set(df_mort.index.get_level_values('Sex')))
        age_list = list(set(df_mort.index.get_level_values('Age')))

        df_list = []
        df_list_future = []
        for sex in sex_list:
            for age in age_list:
                tmp_data = df_mort.loc[(sex, age, slice(None)), :]
                extrap_model = LinearRegression(normalize=True, fit_intercept=True, copy_X=True)
                extrap_model_2 = LinearRegression(normalize=True, fit_intercept=True, copy_X=True)

                first_extrap_df = tmp_data[tmp_data['Years'].between(interval_fit[0], interval_fit[1])]
                XX = tmp_data[tmp_data['Years'].between(interval_fit[0], predict_horizon)].values[:, 0]

                values = first_extrap_df.values
                extrap_model.fit(values[:, 0].reshape(-1, 1), values[:, 1])

                extrap_predictions = extrap_model.predict(XX.reshape(-1, 1))

                loc_df = pd.DataFrame.from_dict({'Sex': sex, 'Age': age, 'Years': XX, 'Extrap': extrap_predictions})
                loc_df.set_index(['Sex', 'Age', 'Years'], inplace=True)

                df_list.append(loc_df.copy())

        df_e1 = pd.concat(df_list, axis=0)

        df_list_final = [df_mort, df_e1]
        df_total = reduce(lambda left, right: pd.merge(left, right, on=['Sex', 'Age', 'Years']), df_list_final)

        df_total = df_total.reset_index(inplace=False).set_index(['Sex', 'Age'], inplace=False)

        df_total['Extrap'] = df_total['Extrap'].apply(np.exp)
        df_total['Data'] = df_total['log_Mortality_mid'].apply(np.exp)
        df_before_time['Data'] = df_before_time['log_Mortality_mid'].apply(np.exp)

        df_before_time.set_index(['Sex', 'Age'], inplace=True)
        df_total = pd.concat([df_total, df_before_time], axis=0, join='outer', sort=True)

        df_total.reset_index(inplace=True)
        df_total['sortby2'] = df_total['Age'].map(age_order_dict)
        df_total['sortby1'] = df_total['Sex'].map(sex_dict)
        df_total.sort_values(by=['sortby1', 'sortby2', 'Years'], inplace=True)
        df_total.drop(columns=['sortby1', 'sortby2'], inplace=True)

        estimates_list = []
        if not add_to_list:
            estimates_list = [df_total.copy()]
        else:
            estimates_list.append(df_total.copy()) 
       
        def min_not_nan(x_list):
            loc_in = list(filter(lambda x: not np.isnan(x), x_list))
            return np.min(loc_in)

        # This was in another function before
        df = estimates_list[n]
        df['FE'] = df[['Data', 'Extrap']].apply(min_not_nan, axis=1)
        df['Age'] = df['Age'].apply(lambda x: int(x.split(',')[1].split(')')[0]))
        male_df = df[df['Sex'] == 'Male']
        female_df = df[df['Sex'] == 'Female']

        male_df.set_index(['Sex','Age', 'Years'], inplace=True)
        female_df.set_index(['Sex','Age','Years'], inplace=True)
        male_data = male_df['FE']
        female_data = female_df['FE']

        male_data = male_data.unstack(-1)
        male_data.sort_index(level ='Age', inplace=True)
        female_data = female_data.unstack(-1)
        female_data.sort_index(level='Age', inplace=True)

        years_out_male = list(male_data.columns)
        years_out_female =list( female_data.columns)

        age_out_male = list(male_data.index.get_level_values('Age'))
        age_out_female = list(male_data.index.get_level_values('Age'))

        male_output= male_data.values
        female_output = female_data.values

        if csv_out:
            male_data.to_csv(f'Male{csv_out}')
            female_data.to_csv(f'Female{csv_out}')

        # TBD: This is the part that should use base file functionality
        
        dict_female = {'NumPopulationGroups': list(female_data.shape),
                       'AxisNames': ['age', 'year'],
                       'AxisScaleFactors': [365.0, 1],
                       'AxisUnits': ['years', 'years'],
                       'NumDistributionAxes': 2,
                       'PopulationGroups': [age_out_female, years_out_female],
                       'ResultScaleFactor': results_scale_factor,
                       'ResultUnits': 'annual deaths per capita',
                       'ResultValues': female_output.tolist()
                       }

        dict_male =   {'NumPopulationGroups': list(male_data.shape),
                       'AxisNames': ['age', 'year'],
                       'AxisScaleFactors': [365.0, 1],
                       'AxisUnits': ['years', 'years'],
                       'NumDistributionAxes': 2,
                       'PopulationGroups': [age_out_male, years_out_male],
                       'ResultScaleFactor': results_scale_factor,
                       'ResultUnits': 'annual deaths per capita',
                       'ResultValues': male_output.tolist()
                       }
        male_mort = { "MortalityDistributionMale" : dict_male }
        female_mort = { "MortalityDistributionFemale" : dict_female }
        self.SetDefaultFromTemplate( male_mort, DT._set_mortality_age_gender_year )
        self.SetDefaultFromTemplate( female_mort )

    def apply_assortivity( self, category, rowmale_colfemale_matrix ):
        """
        E.g., create:
        {
          "Defaults": {
            "Society": {
              "COMMERCIAL": {
                "Pair_Formation_Parameters": {
                  "Assortivity": {
                    "Axes": [ "LOW", "MEDIUM", "HIGH" ],
                    "Group": "INDIVIDUAL_PROPERTY",
                    "Property_Name": "Risk",
                    "Weighting_Matrix_RowMale_ColumnFemale": [
                      [ 1, 1, 1 ],
                      [ 1, 1, 1 ],
                      [ 1, 1, 1 ]
                    ]
                  }
                }
              },
              "INFORMAL": {
                "Pair_Formation_Parameters": {
                  "Assortivity": {
                    "Axes": [ "LOW", "MEDIUM", "HIGH" ],
                    "Group": "INDIVIDUAL_PROPERTY",
                    "Property_Name": "Risk",
                    "Weighting_Matrix_RowMale_ColumnFemale": [
                      [ 0.6097767084, 0.3902232916, 0 ],
                      [ 0.3902232916, 0.6097767084, 0.6097767084 ],
                      [ 0, 0.6097767084, 0.3902232916 ]
                    ],
                  }
                }
              },
              "MARITAL": {
                "Pair_Formation_Parameters": {
                  "Assortivity": {
                    "Axes": [ "LOW", "MEDIUM", "HIGH" ],
                    "Group": "INDIVIDUAL_PROPERTY",
                    "Property_Name": "Risk",
                    "Weighting_Matrix_RowMale_ColumnFemale": [
                      [ 0.6097767084, 0.3902232916, 0 ],
                      [ 0.3902232916, 0.6097767084, 0.6097767084 ],
                      [ 0, 0.6097767084, 0.3902232916 ]
                    ],
                  }
                }
              },
              "TRANSITORY": {
                "Pair_Formation_Parameters": {
                  "Assortivity": {
                    "Axes": [ "LOW", "MEDIUM", "HIGH" ],
                    "Group": "INDIVIDUAL_PROPERTY",
                    "Property_Name": "Risk",
                    "Weighting_Matrix_RowMale_ColumnFemale": [
                      [ 0.6097767084, 0.3902232916, 0 ],
                      [ 0.3902232916, 0.6097767084, 0.6097767084 ],
                      [ 0, 0.6097767084, 0.3902232916 ]
                    ] 
                  }
                }
              }
            }
          }
        }
        """

        if len(rowmale_colfemale_matrix) != 3:
            raise ValueError( "You need to have 3 rows in your assortivity matrix." )
        if any( len(row) != 3 for row in rowmale_colfemale_matrix ):
            raise ValueError( "You need to have 3 columns in each row of your assortivity matrix." )
        assortivity = {}
        assortivity["Axes"] = [ "LOW", "MEDIUM", "HIGH" ]
        assortivity["Group"] = "INDIVIDUAL_PROPERTY"
        assortivity["Property_Name"] = "Risk"
        assortivity["Weighting_Matrix_RowMale_ColumnFemale" ] = rowmale_colfemale_matrix
        self.raw['Defaults']["Society"][category]["Pair_Formation_Parameters"]["Assortivity"].update( assortivity )


def from_template_node(lat=0, lon=0, pop=1e6, name=1, forced_id=1 ):
    """
    Create a single-node :py:class:`~emodpy_hiv.demographics.HIVDemographics`
    instance from the parameters you supply.

    Args:
        lat: Latitude of the centroid of the node to create.
        lon: Longitude of the centroid of the node to create.
        pop: Human population of the node. 
        name: The name of the node. This may be a characteristic of the 
            node, such as "rural" or "urban", or an identifying integer.
        forced_id: The node ID for the single node.

    Returns:
        A :py:class:`~emodpy_malaria.demographics.HIVDemographics` instance.
    """
    new_nodes = [ Demog.Node(lat=lat, lon=lon, pop=pop, name=name, forced_id=forced_id) ] 
    return HIVDemographics(nodes=new_nodes)

def from_pop_csv( pop_filename_in, pop_filename_out="spatial_gridded_pop_dir", site="No_Site" ):
    """
    Create a multi-node :py:class:`~emodpy_hiv.demographics.HIVDemographics`
    instance from a CSV file describing a population.

    Args:
        pop_filename_in: The path to the demographics file to ingest.
        pop_filename_out: The path to the file to output.
        site: A string to identify the country, village, or trial site.

    Returns:
        A :py:class:`~emodpy_malaria.demographics.HIVDemographics` instance.
    """
    generic_demog = Demog.from_pop_csv( pop_filename_in, pop_filename_out, site )
    nodes = generic_demog.nodes
    return HIVDemographics(nodes=nodes, idref=site)

def from_params(tot_pop=1e6, num_nodes=100, frac_rural=0.3, id_ref="from_params" ):
    """
    Create a multi-node :py:class:`~emodpy_hiv.demographics.HIVDemographics`
    instance as a synthetic population based on a few parameters.

    Args:
        tot_pop: The total human population in the node.
        num_nodes: The number of nodes to create.
        frac_rural: The fraction of the population that is rural.
        id_ref: Method describing how the latitude and longitude values are created
            for each of the nodes in a simulation. "Gridded world" values use a grid 
            overlaid across the globe at some arcsec resolution. You may also generate 
            the grid using another tool or coordinate system. For more information,
            see :ref:`emod-hiv:demo-metadata`.

    Returns:
        A :py:class:`~emodpy_hiv.demographics.HIVDemographics` instance.
    """
    generic_demog = Demog.from_params(tot_pop, num_nodes, frac_rural, id_ref )
    nodes = generic_demog.nodes
    return HIVDemographics(nodes=nodes, idref=id_ref )
